"""Test the usability of the C++ extension module and pluggable MD extension code."""

import logging
import os

import pytest

try:
    import gmxapi as gmx
    from gmxapi.simulation.context import Context as _context
    from gmxapi.simulation.workflow import WorkElement, from_tpr
    from gmxapi.version import api_is_at_least
except (ImportError, ModuleNotFoundError):
    import gmx
    from gmx import get_context as _context
    from gmx.version import api_is_at_least
    from gmx.workflow import from_tpr, WorkElement

nompi = lambda f: f
try:
    from mpi4py import MPI
    if MPI.Is_initialized():
        rank = MPI.COMM_WORLD.Get_rank()
        if MPI.COMM_WORLD.Get_size() > 1:
            nompi = pytest.mark.skip(reason='Test cannot run in a multirank MPI environment.')
    else:
        rank = 0

    # Get a fixture for tests that should only run with 2 MPI ranks.
    withmpi_only = pytest.mark.skipif(
        not MPI.Is_initialized() or MPI.COMM_WORLD.Get_size() < 2,
        reason="Test requires at least 2 MPI ranks, but MPI is not initialized or too small.")
except (ImportError, ModuleNotFoundError):
    withmpi_only = pytest.mark.skip(
        reason="Test requires at least 2 MPI ranks, but mpi4py is not available.")
    rank = ''

logging.getLogger().setLevel(logging.DEBUG)
# create console handler
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
# create formatter and add it to the handler
formatter = logging.Formatter(f'%(asctime)s:%(name)s:%(levelname)s:{rank} %(message)s')
ch.setFormatter(formatter)
# add the handlers to the logger
logging.getLogger().addHandler(ch)

logger = logging.getLogger()


def test_import():
    # Suppress inspection warning outside of testing context.
    # noinspection PyUnresolvedReferences
    import brer
    # TODO: Consider the new "lazy import" scheme as an option to make cleaner
    #  static checking of the following line.
    import brer.md
    assert callable(brer.md.brer_restraint)



"""BRER data for site pairs.

Coordinate the experimental reference data and molecular model data for the
labeled / restrained pairs.

Support the statistical (re)sampling of target pair distances when beginning a
BRER iteration.
"""
import dataclasses
import json
import os
import pathlib
import typing
import warnings
from typing import Iterator

import numpy as np

from ._compat import dataclass_kw_only
from ._compat import List
from ._compat import Mapping


@dataclasses.dataclass(frozen=True)
class PairData:
    """Pair distance distribution.

    Essential pair data to support BRER MD plugin code. All data must be
    provided when the object is initialized. (Fields are read-only.)

    Fields here correspond to the fields for each named pair
    (JSON *objects*) in a :file:`pair_data.json` file (the *pairs_json* argument
    of :py:func:`~brer.run_config.RunConfig()`).

    .. versionchanged:: 2.0

        When reading from
        :file:`pair_data.json` or :file:`state.json`, the *name* used for
        PairData comes from the object key, not from the *name* field of the object.
        The *name* field in the serialized (JSON) representation is ignored
        when reading, but is preserved for backward compatibility when writing.

    """
    # Historically, *name* has sometimes been provided as a positional argument,
    # but this led to ambiguity in the source of the final value of the field.
    # We define the *name* field first in case we overlook some legacy usage.
    name: str = dataclasses.field(**dataclass_kw_only)
    """Identifier for the pair of sites on the molecule.
    
    This string is chosen by the researcher. For example, the name may
    include identifiers for the two residues in a scheme that can be easily
    cross-referenced with experimental data.
    """

    bins: List[float] = dataclasses.field(**dataclass_kw_only)
    """Histogram edges for the distance distribution data.
    
    (Simulation length units.)
    """

    distribution: List[float] = dataclasses.field(**dataclass_kw_only)
    """Site distance distribution.
    
    Histogram values (weights or relative probabilities) for distances between
    the sites. (Generally derived from experimental data.)
    """

    sites: List[int] = dataclasses.field(**dataclass_kw_only)
    """Indices defining the distance vector.
    
    A list of indices for sites in the molecular model. The first and last
    list elements are the sites associated with the distance data. Additional
    indices can be inserted in the list to define a chain of distance vectors
    that will be added without applying periodic boundary conditions.
    
    If, at any point in the simulation, the two molecular sites in the pair
    might be farther apart than half of the shortest simulation box dimension,
    the distance might accidentally get calculated between sites on different
    molecule "images" (periodic boundary conditions). To make sure that site-site
    distances are calculated on the same molecule, provide a sequence of sites
    on the molecule (that are never more than half a box-length apart) so that
    the correct vector between sites is unambiguous.
    """


class PairDataCollection(Mapping[str, PairData]):
    """Data for all the restrained pairs in a BRER simulation.

    Source data for the pair restraints is provided through a JSON file
    (*pairs_json*). The JSON file contains one *JSON object* for each
    `PairData` to be read.

    For each object, the object key is assumed to be the
    `PairData.name` of a pair.
    The JSON object contents are used to initialize a
    `PairData` for each named pair.

    The data file is usually constructed manually by the researcher after
    inspection of a molecular model and available experimental data. An example
    of what such a file should look like is provided in the :file:`brer/data`
    directory of the installed package or
    `in the source <https://github.com/kassonlab/brer-md/tree/main/src/brer/data>`__
    repository.

    Note that JSON is not a Python-specific file format, but
    :py:mod:`json` may be helpful.

    A *PairDataCollection* can be initialized from a sequence of `PairData`
    objects, or created from a JSON pair data file by using `create_from()`.

    """

    def __init__(self, *pairs: PairData):
        if len(pairs) < 1 or not all(isinstance(pair, PairData) for pair in pairs):
            raise TypeError('Must be initialized with one or more PairData instances.')
        self._pairs = {pair.name: pair for pair in pairs}

    def __getitem__(self, key: str) -> PairData:
        return self._pairs[key]

    def __iter__(self) -> Iterator[str]:
        yield from self._pairs

    def __len__(self) -> int:
        return len(self._pairs)

    @staticmethod
    def create_from(filename: typing.Union[str, os.PathLike, pathlib.Path]):
        """Reads pair data from json file.

        Parameters
        ----------
        filename :
            filename of the pair data


        """
        pairs = []
        with open(filename, 'r') as fh:
            for name, obj in json.load(fh).items():
                kwargs = {}
                for key, value in obj.items():
                    if key == 'name':
                        message = '*name* field is ignored from brer-md 2.0.'
                        if name == value:
                            warnings.warn(message, DeprecationWarning)
                        else:
                            message += f' Using {name} instead of {value}.'
                            warnings.warn(message, UserWarning)
                    else:
                        kwargs[key] = value
                pairs.append(PairData(name=name, **kwargs))
        return PairDataCollection(*pairs)

    def as_dict(self):
        """Encode the full collection as a single Python dictionary."""
        return {pair.name: dataclasses.asdict(pair) for pair in self._pairs.values()}


def sample(pair_data: PairData):
    """Choose a bin edge according to the probability distribution."""
    distribution = pair_data.distribution
    bins = pair_data.bins

    normalized = np.divide(distribution, np.sum(distribution))
    return np.random.choice(bins, p=normalized)


def sample_all(pairs: PairDataCollection):
    """Get a mapping of pair names to freshly sampled targets."""
    return {name: sample(pair) for name, pair in pairs.items()}



"""Configuration and fixtures for pytest."""

import json
import logging
import os
import shutil
import sys
import tempfile
import warnings
from contextlib import contextmanager

import pytest

if sys.version_info.major > 3 or sys.version_info.minor >= 10:
    from importlib.resources import files, as_file
else:
    from importlib_resources import files, as_file


@pytest.fixture(scope='session')
def simulation_input():
    source = files('brer').joinpath('data', 'topol.tpr')
    with as_file(source) as tpr_file:
        yield tpr_file


@pytest.fixture(scope='session')
def pair_data_file():
    source = files('brer').joinpath('data', 'pair_data.json')
    with as_file(source) as pair_data:
        yield pair_data


@pytest.fixture(scope='session')
def raw_pair_data():
    """
    Three DEER distributions for testing purposes.

    :return: contents of :file:`brer/data/pair_data.json`
    """
    raw_data = files('brer').joinpath('data', 'pair_data.json').read_text()
    pair_data = json.loads(raw_data)
    assert pair_data["196_228"]["distribution"][0] == 3.0993964770242886e-55
    return pair_data

# TODO: This issue should be resolved. Remove the monkey-patch.
# Need to monkey-patch subprocess.run until gmxapi can be fixed.
# MPI-related environment variables will cause GROMACS command line tools to
# try to MPI_Init_thread (and fail) even though mpi4py has already called MPI_Init.
import functools
import subprocess
env = {key: value for key, value in os.environ.items() if key.startswith('GMX')}
env['PATH'] = os.getenv('PATH')
subprocess.run = functools.partial(subprocess.run, env=env)


def pytest_addoption(parser):
    """Add a command-line user option for the pytest invocation."""
    parser.addoption(
        '--rm',
        action='store',
        default='always',
        choices=['always', 'never', 'success'],
        help='Remove temporary directories "always", "never", or on "success".'
    )


@pytest.fixture(scope='session')
def remove_tempdir(request):
    """pytest fixture to get access to the --rm CLI option."""
    return request.config.getoption('--rm')


@contextmanager
def scoped_chdir(dir):
    oldpath = os.getcwd()
    os.chdir(dir)
    try:
        yield dir
        # If the `with` block using scoped_chdir produces an exception, it will
        # be raised at this point in this function. We want the exception to
        # propagate out of the `with` block, but first we want to restore the
        # original working directory, so we skip `except` but provide a `finally`.
    finally:
        os.chdir(oldpath)


@contextmanager
def _cleandir(remove_tempdir):
    """Context manager for a clean temporary working directory.

    Arguments:
        remove_tempdir (str): whether to remove temporary directory "always",
                              "never", or on "success"

    The context manager will issue a warning for each temporary directory that
    is not removed.
    """

    newpath = tempfile.mkdtemp()

    def remove():
        shutil.rmtree(newpath)

    def warn():
        warnings.warn('Temporary directory not removed: {}'.format(newpath))

    if remove_tempdir == 'always':
        callback = remove
    else:
        callback = warn
    try:
        with scoped_chdir(newpath):
            yield newpath
        # If we get to this line, the `with` block using _cleandir did not throw.
        # Clean up the temporary directory unless the user specified `--rm never`.
        # I.e. If the user specified `--rm success`, then we need to toggle from `warn` to `remove`.
        if remove_tempdir != 'never':
            callback = remove
    finally:
        callback()


@pytest.fixture
def cleandir(remove_tempdir):
    """Provide a clean temporary working directory for a test.

    Example usage:

        import os
        import pytest

        @pytest.mark.usefixtures("cleandir")
        def test_cwd_starts_empty():
            assert os.listdir(os.getcwd()) == []
            with open("myfile", "w") as f:
                f.write("hello")

        def test_cwd_also_starts_empty(cleandir):
            assert os.listdir(os.getcwd()) == []
            assert os.path.abspath(os.getcwd()) == os.path.abspath(cleandir)
            with open("myfile", "w") as f:
                f.write("hello")

        @pytest.mark.usefixtures("cleandir")
        class TestDirectoryInit(object):
            def test_cwd_starts_empty(self):
                assert os.listdir(os.getcwd()) == []
                with open("myfile", "w") as f:
                    f.write("hello")

            def test_cwd_also_starts_empty(self):
                assert os.listdir(os.getcwd()) == []
                with open("myfile", "w") as f:
                    f.write("hello")

    Ref: https://docs.pytest.org/en/latest/fixture.html#using-fixtures-from-classes-modules-or-projects
    """
    with _cleandir(remove_tempdir) as newdir:
        yield newdir


@pytest.fixture(scope='session')
def gmxcli():
    try:
        import gmxapi.commandline
        command = gmxapi.commandline.cli_executable()
    except (ImportError, AttributeError):
        # The gmxapi version predates the cli_executable() utility.
        # Search for the cli binary.
        allowed_command_names = ['gmx', 'gmx_mpi']
        command = None
        for command_name in allowed_command_names:
            if command is not None:
                break
            command = shutil.which(command_name)
            if command is None:
                gmxbindir = os.getenv('GMXBIN')
                if gmxbindir is None:
                    gromacsdir = os.getenv('GROMACS_DIR')
                    if gromacsdir is not None and gromacsdir != '':
                        gmxbindir = os.path.join(gromacsdir, 'bin')
                if gmxbindir is None:
                    gmxapidir = os.getenv('gmxapi_DIR')
                    if gmxapidir is not None and gmxapidir != '':
                        gmxbindir = os.path.join(gmxapidir, 'bin')
                if gmxbindir is not None:
                    gmxbindir = os.path.abspath(gmxbindir)
                    command = shutil.which(command_name, path=gmxbindir)
    if command is None:
        message = "Tests need 'gmx' command line tool, but could not find it on the path."
        raise RuntimeError(message)
    try:
        assert os.access(command, os.X_OK)
    except Exception as E:
        raise RuntimeError('"{}" is not an executable gmx wrapper program'.format(command)) from E
    yield str(command)


@pytest.fixture(scope='class')
def spc_water_box(gmxcli, remove_tempdir):
    """Provide a TPR input file for a simple simulation.

    Prepare the MD input in a freshly created working directory.
    """
    try:
        import gmxapi as gmx
    except (ImportError, ModuleNotFoundError):
        current_dir = os.path.dirname(__file__)
        file_path = os.path.join(current_dir, 'data', 'topol.tpr')
        yield os.path.abspath(file_path)
    else:
        # TODO: (#2896) Fetch MD input from package / library data.
        # Example:
        #     import pkg_resources
        #     # Note: importing pkg_resources means setuptools is required for running this test.
        #     # Get or build TPR file from data bundled via setup(package_data=...)
        #     # Ref https://setuptools.readthedocs.io/en/latest/setuptools.html#including-data-files
        #     from gmx.data import tprfilename

        with _cleandir(remove_tempdir) as tempdir:

            testdir = os.path.dirname(__file__)
            with open(os.path.join(testdir, 'testdata.json'), 'r') as fh:
                testdata = json.load(fh)

            # TODO: (#2756) Don't rely on so many automagical behaviors (as described in comments below)

            structurefile = os.path.join(tempdir, 'structure.gro')
            # We let `gmx solvate` use the default solvent. Otherwise, we would do
            #     gro_input = testdata['solvent_structure']
            #     with open(structurefile, 'w') as fh:
            #         fh.write('\n'.join(gro_input))
            #         fh.write('\n')

            topfile = os.path.join(tempdir, 'topology.top')
            top_input = testdata['solvent_topology']
            # `gmx solvate` will append a line to the provided file with the molecule count,
            # so we strip the last line from the input topology.
            with open(topfile, 'w') as fh:
                fh.write('\n'.join(top_input[:-1]))
                fh.write('\n')

            assert os.path.exists(topfile)
            solvate = gmx.commandline_operation(gmxcli,
                                                arguments=['solvate', '-box', '5', '5', '5'],
                                                # We use the default solvent instead of specifying one.
                                                # input_files={'-cs': structurefile},
                                                output_files={'-p': topfile,
                                                              '-o': structurefile,
                                                              }
                                                )
            if not os.path.exists(topfile):
                raise RuntimeError(f'{topfile} does not exist.')

            logging.debug('Running solvate.')
            if solvate.output.returncode.result() != 0:
                logging.debug('Solvate error output: ' + solvate.output.erroroutput.result())
                raise RuntimeError('solvate failed in spc_water_box testing fixture.')

            # Choose an exactly representable dt of 2^-9 ps (approximately 0.002)
            dt = 2.**-9.
            mdp_input = [('integrator', 'md'),
                         ('dt', dt),
                         ('cutoff-scheme', 'Verlet'),
                         ('nsteps', 2),
                         ('nstxout', 1),
                         ('nstvout', 1),
                         ('nstfout', 1),
                         ('tcoupl', 'v-rescale'),
                         ('tc-grps', 'System'),
                         ('tau-t', 1),
                         ('ref-t', 298)]
            mdp_input = '\n'.join([' = '.join([str(item) for item in kvpair]) for kvpair in mdp_input])
            mdpfile = os.path.join(tempdir, 'md.mdp')
            with open(mdpfile, 'w') as fh:
                fh.write(mdp_input)
                fh.write('\n')
            tprfile = os.path.join(tempdir, 'topol.tpr')
            # We don't use mdout_mdp, but if we don't specify it to grompp,
            # it will be created in the current working directory.
            mdout_mdp = os.path.join(tempdir, 'mdout.mdp')

            grompp = gmx.commandline_operation(gmxcli, 'grompp',
                                               input_files={
                                                   '-f': mdpfile,
                                                   '-p': solvate.output.file['-p'],
                                                   '-c': solvate.output.file['-o'],
                                                   '-po': mdout_mdp,
                                               },
                                               output_files={'-o': tprfile})
            tprfilename = grompp.output.file['-o'].result()
            if grompp.output.returncode.result() != 0:
                logging.debug(grompp.output.erroroutput.result())
                raise RuntimeError('grompp failed in spc_water_box testing fixture.')

            # TODO: more inspection of grompp errors...
            if not os.path.exists(tprfilename):
                raise RuntimeError(f'Failed to produce {tprfilename}')

            yield tprfilename



"""Unit tests and regression for DirectoryHelper class."""
import os

from brer.directory_helper import DirectoryHelper


def test_directory(tmpdir):
    """Checks that directory creation for BRER runs is functional.

    Parameters
    ----------
    tmpdir :
        temporary pytest directory.
    """
    my_home = os.path.abspath(os.getcwd())
    top_dir = tmpdir.mkdir("top_directory")
    dir_helper = DirectoryHelper(top_dir,
                                 {'ensemble_num': 1, 'iteration': 0, 'phase': 'training'})
    dir_helper.build_working_dir()
    dir_helper.change_dir('phase')
    assert (os.getcwd() == '{}/mem_{}/{}/{}'.format(top_dir, 1, 0, 'training'))

    os.chdir(my_home)



"""RunConfig class handles the actual workflow logic."""
__all__ = ("RunConfig",)

import collections.abc
import dataclasses
import functools
import importlib
import logging
import os
import pathlib
import shutil
import sys
import typing
import warnings
from typing import Sequence
from typing import Union

from .pair_data import PairDataCollection
from .pair_data import sample_all

from .directory_helper import DirectoryHelper
from .plugin_configs import ConvergencePluginConfig
from .plugin_configs import PluginConfig
from .plugin_configs import ProductionPluginConfig
from .plugin_configs import TrainingPluginConfig
from .run_data import RunData

_Path = Union[str, pathlib.Path]


def _gmxapi_missing(*args, msg: str = "", **kwargs):
    """Placeholder function for missing gmxapi functionality.

    Allows import errors to be deferred until run time to aid in docs builds
    and troubleshooting. Try to provide a useful RuntimeError that includes
    the details of the import error(s).
    """
    _message = (
        "brer requires gmxapi. See https://github.com/kassonlab/brer_md#requirements"
    )
    if msg:
        _message = "\n".join((_message, msg))
    raise RuntimeError(_message)


def get_api_callable(attr: str, modules: typing.Iterable[str]):
    """Get a gmxapi callable or placeholder.

    Try to import *attr* from successive *modules*. Return the first callable
    found, else return a placeholder that emits a RuntimeError when called.
    """
    message = ""
    version = ""
    func = None
    for module in modules:
        try:
            mod = importlib.import_module(module)
            func = getattr(mod, attr)
            base = module.split(".")[0]
            version = sys.modules[base].__version__
        except ImportError as e:
            message = "\n".join((message, f"Could not import {module}: {str(e)}"))
        else:
            break
    if callable(func):
        qualname = ".".join((func.__module__, func.__name__))
    else:
        func = functools.partial(_gmxapi_missing, msg=message)
        qualname = ".".join((_gmxapi_missing.__module__, "_gmxapi_missing"))

    report = "Using" \
             + " ".join((qualname, version)) \
             + " for {attr}."
    logging.info(report)
    return func


_context = get_api_callable("Context", ("gmxapi.simulation.context", "gmx.context"))
from_tpr = get_api_callable("from_tpr", ("gmxapi.simulation.workflow", "gmx.workflow"))
WorkElement = get_api_callable("WorkElement", ("gmxapi.simulation.workflow", "gmx.workflow"))


def check_consistency(*, data: PairDataCollection, state: RunData):
    """Check for mismatched data sources.

    Returns
    -------
    bool
        True if pairs match across sources, False otherwise.

    """
    if set(data.keys()) != set(state.pair_params.keys()):
        return False
    for name in data:
        if data[name].sites != state.pair_params[name].sites:
            return False
    return True


def _get_communicator(ensemble_size: int):
    try:
        import mpi4py.MPI as _MPI
    except (ImportError, ModuleNotFoundError):
        _MPI = None
    if _MPI is None or _MPI.COMM_WORLD.Get_size() < ensemble_size:
        raise RuntimeError('Need mpi4py and one MPI rank per ensemble member.')
    else:
        communicator: _MPI.Comm = _MPI.COMM_WORLD
        assert communicator.Get_size() >= ensemble_size
        # TODO: Handle mismatched ensemble size.
        if communicator.Get_size() > ensemble_size:
            warnings.warn('brer does not yet attempt to handle communicators '
                          'larger than the ensemble.')
        return communicator


class RunConfig:
    """Run configuration for single BRER ensemble member.

    The run configuration specifies the files and directory structure
    used for the run. It determines whether the run is in the training,
    convergence, or production phase, then performs the run.

    Note that all instances of RunConfig need the same sized array of TPR input
    files across all ranks in an MPI ensemble because they must all be capable of
    constructing a compatible copy of the ensemble simulation work description.

    Source data for the pair restraints is provided through a JSON file
    (*pairs_json*). See :py:class:`brer.pair_data.PairDataCollection` for details.

    Parameters
    ----------
    tpr : str
        path (or paths) to tpr input. Must be compatible with the GROMACS version
        providing gmxapi.
    ensemble_dir : str
        path to top directory which contains the full ensemble.
    ensemble_num : int, default=0
        the ensemble member to run
    pairs_json : str, default="pair_data.json"
        path to file containing *ALL* the pair metadata.
        (A serialized :py:class:`brer.pair_data.PairDataCollection`.)


    """

    @property
    def tpr(self):
        return pathlib.Path(self._tprs[self._rank])

    def __init__(self,
                 tpr: Union[_Path, Sequence[_Path]],
                 ensemble_dir,
                 ensemble_num: int = None,
                 pairs_json='pair_data.json'):
        if isinstance(tpr, (str, pathlib.Path, os.PathLike)):
            self._tprs = tuple([str(tpr)])
            self._ensemble_size = 1
        else:
            if not isinstance(tpr, collections.abc.Iterable):
                raise ValueError('Paramater *tpr* must be an input file or sequence of '
                                 'input files (for ensemble input).')
            self._tprs = tuple([str(name) for name in tpr])
            self._ensemble_size = len(self._tprs)
        if self._ensemble_size == 1:
            self._communicator = None
            self._rank = 0
        else:
            self._communicator = _get_communicator(self._ensemble_size)
            self._rank = self._communicator.Get_rank()

        if ensemble_num is None:
            ensemble_num = self._rank
        else:
            if self._communicator is not None:
                # Greater future flexibility is described at
                # https://github.com/kassonlab/brer-md/issues/7
                raise TypeError(
                    'RunConfig does not allow *ensemble_num* with mpi4py ensembles.')

        if not os.path.exists(ensemble_dir):
            raise RuntimeError(f'Ensemble directory {ensemble_dir} does not exist!')
        self.ens_dir = ensemble_dir

        # Load the pair data from a json. Use this to set up the run metadata
        self.pairs = PairDataCollection.create_from(pairs_json)
        # use the same identifiers for the pairs here as those provided in the pair
        # metadata
        # file this prevents mixing up pair data amongst the different pairs (i.e.,
        # accidentally applying the restraints for pair 1 to pair 2.)
        self.__names = tuple(name for name in self.pairs)
        """A list of identifiers of the residue-residue pairs that will be restrained."""

        member_directory = os.path.join(self.ens_dir, f'mem_{ensemble_num}')
        if not os.path.exists(member_directory):
            os.mkdir(member_directory)

        self.state_json = '{}/mem_{}/state.json'.format(ensemble_dir,
                                                        ensemble_num)
        # If we're in the middle of a run, load the BRER checkpoint file and continue from
        # the current state.
        if os.path.exists(self.state_json):
            self.run_data = RunData.create_from(self.state_json, ensemble_num=ensemble_num)
            if not check_consistency(data=self.pairs, state=self.run_data):
                raise RuntimeError('Checkpoint data inconsistent with pair data source.')
        # Otherwise, populate the state information using the provided pair data
        # and default parameter values.
        # Then save the current state.
        else:
            self.run_data = RunData.create_from(self.pairs, ensemble_num=ensemble_num)
            self.run_data.save_config(self.state_json)

        # List of plugins
        self.__plugins = []

        # Logging
        self._logger = logging.getLogger('BRER')
        self._logger.setLevel(logging.DEBUG)
        # create file handler which logs even debug messages
        fh = logging.FileHandler('brer{}.log'.format(ensemble_num))
        fh.setLevel(logging.DEBUG)
        # create console handler with a higher log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        # create formatter and add it to the handlers
        formatter = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        # add the handlers to the logger
        self._logger.addHandler(fh)
        self._logger.addHandler(ch)

        self._logger.info("Initialized the run configuration: {}".format(
            self.run_data.as_dictionary()))
        self._logger.info("Names of restraints: {}".format(self.__names))

        # Need to cleanly handle cancelled jobs: write out a checkpoint of the state if
        # the
        # job is exited.
        # def cleanup():
        #     """"""
        #     self.run_data.save_config(self.state_json)
        #     self._logger.info("BRER received INT signal, stopping and saving data to
        #     {}".format(self.state_json))

        # atexit.register(cleanup)

    def build_plugins(self, plugin_config: typing.Type[PluginConfig]):
        """Builds the plugin configuration.

        For each pair-wise restraint,
        populate the plugin with data: both the "general" data and the data
        unique to that restraint.

        Parameters
        ----------
        plugin_config : PluginConfig
            the particular plugin configuration (Training, Convergence, Production) for
            the run.
        """

        # One plugin per restraint.
        self.__plugins = []
        # For each pair-wise restraint, populate the plugin with data: both the
        # "general" data and
        # the data unique to that restraint.
        for name in self.__names:
            pair_params = dataclasses.asdict(self.run_data.general_params)
            pair_params.update(dataclasses.asdict(self.run_data.pair_params[name]))
            new_restraint = plugin_config.create_from(pair_params)
            self.__plugins.append(new_restraint.build_plugin())

    def __set_workdir(self):
        # change into the current working directory (
        # ensemble_path/member_path/iteration/phase)
        dir_help = DirectoryHelper(
            top_dir=self.ens_dir,
            param_dict=dataclasses.asdict(self.run_data.general_params))
        dir_help.build_working_dir()
        workdir = dir_help.change_dir('phase')
        if self._communicator is None or self._communicator.Get_size() == 1:
            self.workdirs = [workdir]
        else:
            # assert isinstance(self._communicator, MPI.Comm)
            comm_size: int = self._communicator.Get_size()
            assert comm_size > 1
            assert self._rank < comm_size
            self.workdirs = self._communicator.allgather(workdir)
            assert len(self.workdirs) == comm_size
            assert self.workdirs[self._rank] == workdir

    def __move_cpt(self):

        def safe_copy(src, dst):
            if not os.path.exists(src):
                raise RuntimeError('Missing file: {}'.format(src))
            if os.path.exists(dst):
                raise RuntimeError('Destination file already exists: {}'.format(dst))
            size = os.stat(src).st_size
            if size == 0:
                raise RuntimeError('Source file has zero size: {}'.format(src))
            target_tmp = dst + '.tmp'
            try:
                shutil.copy(src, target_tmp)
            except Exception as e:
                if os.path.exists(target_tmp):
                    os.unlink(target_tmp)
                raise e
            os.rename(target_tmp, dst)
            assert os.stat(dst).st_size > 0

        current_iter = int(self.run_data.get('iteration'))
        ens_num = int(self.run_data.get('ensemble_num'))
        phase: str = self.run_data.get('phase')

        target_dir = os.getcwd()
        # Check logic implicit in prior revisions.
        assert os.path.abspath('{}/mem_{}/{}/{}'.format(self.ens_dir,
                                                        ens_num,
                                                        current_iter,
                                                        phase)) == target_dir

        target = os.path.join(target_dir, 'state.cpt')
        # If the cpt already exists, don't overwrite it
        if os.path.exists(target):
            self._logger.info(
                "Phase is {} and state.cpt already exists: not moving any files".format(
                    phase))

        else:
            member_dir = '{}/mem_{}'.format(self.ens_dir, ens_num)
            prev_iter = current_iter - 1

            if phase in ['training', 'convergence']:
                if prev_iter > -1:
                    # Get the production cpt from previous iteration
                    source = '{}/{}/production/state.cpt'.format(member_dir, prev_iter)
                    if not os.path.exists(source):
                        raise RuntimeError(
                            'Missing checkpoint file from previous iteration: {}'.format(
                                source))
                    safe_copy(source, target)

                else:
                    pass  # Do nothing. Let mdrun generate the initial checkpoint file.

            else:
                # Get the convergence cpt from current iteration
                source = '{}/{}/convergence/state.cpt'.format(member_dir, current_iter)
                if not os.path.exists(source):
                    self._logger.error(f'os.path.exists({source}) is False! Getting '
                                       f'directory listing.')
                    self._logger.error(str(os.listdir(os.path.dirname(source))))
                    raise RuntimeError(
                        'Missing checkpoint file from convergence phase: {}'.format(
                            source))
                safe_copy(source, target)

    def __prep_input(self, tpr_file: str = None):
        if tpr_file is None:
            tpr_file = self.tpr
            # Get the checkpoint file from the previous phase
            self.__move_cpt()
        if not os.path.exists(tpr_file):
            raise RuntimeError('Missing input file: {}'.format(tpr_file))
        return tpr_file

    def __train(self, tpr_file=None, **kwargs):
        for key in ('append_output',):
            if key in kwargs:
                raise TypeError('Conflicting key word argument. Cannot accept {}.'.format(
                    key))

        # do re-sampling
        targets = sample_all(self.pairs)
        self._logger.info('New targets: {}'.format(targets))
        for name in self.__names:
            self.run_data.set(name=name, target=targets[name])

        # save the new targets to the BRER checkpoint file.
        self.run_data.save_config(fnm=self.state_json)

        workdir = self.workdirs[self._rank]

        # backup existing checkpoint.
        # TODO: Don't backup the cpt, actually use it!!
        cpt = '{}/state.cpt'.format(workdir)
        if os.path.exists(cpt):
            self._logger.warning(
                'There is a checkpoint file in your current working directory, but you '
                'are '
                'training. The cpt will be backed up and the run will start over with '
                'new targets')
            shutil.move(cpt, '{}.bak'.format(cpt))

        # If this is not the first BRER iteration, grab the checkpoint from the production
        # phase of the last round
        self.__prep_input(tpr_file)

        # Set up a dictionary to go from plugin name -> restraint name
        sites_to_name = {}

        # Build the gmxapi session.
        tpr_list: Sequence[str] = self._tprs
        md = from_tpr(tpr_list, append_output=False, **kwargs)
        self.build_plugins(TrainingPluginConfig)
        if len(self.__plugins) == 0:
            warnings.warn('No BRER restraints are being applied! User error?')
        for plugin in self.__plugins:
            plugin_name = plugin.name
            for name in self.__names:
                run_data_sites = "{}".format(self.run_data.get('sites', name=name))
                if run_data_sites == plugin_name:
                    sites_to_name[plugin_name] = name
            md.add_dependency(plugin)
        context = _context(md,
                           workdir_list=self.workdirs,
                           communicator=self._communicator)

        self._logger.info("=====TRAINING INFO======\n")
        self._logger.info(f'Working directory: {workdir}')

        # Run it.
        # WARNING: We do not yet handle situations where a rank has no work to do.
        # See https://github.com/kassonlab/run_brer/issues/18
        # and https://github.com/kassonlab/run_brer/issues/55
        with context as session:
            session.run()

        # Through at least gmxapi 0.4, the *potentials* attribute is created on
        # the Context for any Session launched with MD work to perform. An explicit
        # error message here should be more helpful than an AttributeError below,
        # but we don't really know what went wrong.
        # Ref https://github.com/kassonlab/run_brer/issues/55
        if not hasattr(context, 'potentials'):
            raise RuntimeError('Invalid gmxapi Context: missing "potentials" attribute.')

        for i in range(len(self.__names)):
            # noinspection PyUnresolvedReferences
            current_name = sites_to_name[context.potentials[i].name]
            # In the future runs (convergence, production) we need the ABSOLUTE VALUE
            # of alpha.
            # noinspection PyUnresolvedReferences
            current_alpha = context.potentials[i].alpha
            if current_alpha == 0.0:
                raise RuntimeError('Alpha value was constrained to 0.0, which indicates something went wrong')

            # noinspection PyUnresolvedReferences
            current_target = context.potentials[i].target

            self.run_data.set(name=current_name, alpha=current_alpha)
            self.run_data.set(name=current_name, target=current_target)
            self._logger.info("Plugin {}: alpha = {}, target = {}".format(current_name,
                                                                          current_alpha,
                                                                          current_target))

        return context

    def __converge(self, tpr_file=None, **kwargs):

        for key in ('append_output',):
            if key in kwargs:
                raise TypeError('Conflicting key word argument. Cannot accept {}.'.format(
                    key))

        self.__prep_input(tpr_file)

        md = from_tpr(self._tprs, append_output=False, **kwargs)
        self.build_plugins(ConvergencePluginConfig)
        if len(self.__plugins) == 0:
            warnings.warn('No BRER restraints are being applied! User error?')
        for plugin in self.__plugins:
            md.add_dependency(plugin)

        workdir = os.getcwd()
        self._logger.info("=====CONVERGENCE INFO======\n")
        self._logger.info(f'Working directory: {workdir}')

        context = _context(md,
                           workdir_list=self.workdirs,
                           communicator=self._communicator)
        # WARNING: We do not yet handle situations where a rank has no work to do.
        # See https://github.com/kassonlab/run_brer/issues/18
        # and https://github.com/kassonlab/run_brer/issues/55
        with context as session:
            session.run()

        # Through at least gmxapi 0.4, the *potentials* attribute is created on
        # the Context for any Session launched with MD work to perform. An explicit
        # error message here should be more helpful than an AttributeError below,
        # but we don't really know what went wrong.
        # Ref https://github.com/kassonlab/run_brer/issues/55
        if not hasattr(context, 'potentials'):
            raise RuntimeError('Invalid gmxapi Context: missing "potentials" attribute.')

        # Get the absolute time (in ps) at which the convergence run finished.
        # This value will be needed if a production run needs to be restarted.
        # noinspection PyUnresolvedReferences
        self.run_data.set(start_time=context.potentials[0].time)
        for name in self.__names:
            current_alpha = self.run_data.get('alpha', name=name)
            current_target = self.run_data.get('target', name=name)
            message = f'Plugin {name}: alpha = {current_alpha}, target = {current_target}'
            self._logger.info(message)

        return context

    def __production(self, tpr_file=None, **kwargs):

        for key in ('append_output', 'end_time'):
            if key in kwargs:
                raise TypeError('Conflicting key word argument. Cannot accept {}.'.format(
                    key))

        tpr_list = list(self._tprs)
        tpr_list[self._rank] = self.__prep_input(tpr_file)
        if tpr_file is not None:
            # If bootstrap TPR is provided, we are not continuing from the
            # convergence phase trajectory.
            self.run_data.set(start_time=0.0)

        # Calculate the time (in ps) at which the trajectory for this BRER iteration should finish.
        # This should be: the end time of the convergence run + the amount of time for
        # production simulation (specified by the user).
        start_time = self.run_data.get('start_time')
        target_end_time = self.run_data.get('production_time') + start_time

        md = from_tpr(tpr_list, end_time=target_end_time, append_output=False, **kwargs)

        self.build_plugins(ProductionPluginConfig)
        if len(self.__plugins) == 0:
            warnings.warn('No BRER restraints are being applied! User error?')
        for plugin in self.__plugins:
            md.add_dependency(plugin)

        workdir = os.getcwd()
        self._logger.info("=====PRODUCTION INFO======\n")
        self._logger.info(f'Working directory: {workdir}')

        context = _context(md,
                           workdir_list=self.workdirs,
                           communicator=self._communicator
                           )
        # WARNING: We do not yet handle situations where a rank has no work to do.
        # See https://github.com/kassonlab/run_brer/issues/18
        # and https://github.com/kassonlab/run_brer/issues/55
        with context as session:
            session.run()

        # Through at least gmxapi 0.4, the *potentials* attribute is created on
        # the Context for any Session launched with MD work to perform. An explicit
        # error message here should be more helpful than an AttributeError below,
        # but we don't really know what went wrong.
        # Ref https://github.com/kassonlab/run_brer/issues/55
        if not hasattr(context, 'potentials'):
            raise RuntimeError('Invalid gmxapi Context: missing "potentials" attribute.')

        # Get the start and end times for the simulation managed by this Python interpreter.
        # Note that these are the times for all potentials in a single simulation
        # (which should be the same). We are not gathering values across any potential ensembles.
        start_times = [potential.start_time for potential in context.potentials if hasattr(
            potential, 'start_time')]
        if len(start_times) > 0:
            session_start_time = start_times[0]
            if not all(session_start_time == t for t in start_times):
                self._logger.warning('Potentials report inconsistent start times: '
                                     ', '.join(str(t) for t in start_times))
            assert session_start_time >= start_time
        else:
            # If the plugin attribute is missing, assume that the convergence phase behaved properly.
            session_start_time = start_time

        end_times = [potential.time for potential in context.potentials if hasattr(potential, 'time')]
        if len(end_times) > 0:
            session_end_time = end_times[0]
            if not all(session_end_time == t for t in end_times):
                self._logger.warning('Potentials report inconsistent end times: '
                                     ', '.join(str(t) for t in end_times))
        else:
            session_end_time = None

        if session_end_time is not None:
            self.run_data.set(end_time=session_end_time)

        trajectory_time = None
        if session_end_time is not None:
            trajectory_time = session_end_time - session_start_time

        if trajectory_time is not None:
            self._logger.info(f"{trajectory_time} ps production phase trajectory segment.")
        for name in self.__names:
            current_alpha = self.run_data.get('alpha', name=name)
            current_target = self.run_data.get('target', name=name)
            self._logger.info("Plugin {}: alpha = {}, target = {}".format(name,
                                                                          current_alpha,
                                                                          current_target))

        return context

    def run(self, tpr_file=None, **kwargs):
        """Perform the MD simulations.

        Each Python interpreter process runs a separate ensemble member.

        Parameters
        ----------
        tpr_file : str, optional
            If provided, use this input file instead of the input from the main
            configuration.
        **kwargs : optional
            Additional key word arguments are passed on to the simulator.


        After the first "iteration", brer bootstraps the training and convergence
        phase's trajectory with the checkpoint file from the previous iteration's
        production phase.

        At the beginning of a production phase (when there is not yet a checkpoint file),
        the checkpoint file from the convergence phase is used to start the production
        trajectory **unless** *tpr_file* is given.

        When *tpr_file* is not None, run() does not look for a bootstrapping checkpoint
        file. This can be helpful if a checkpoint file is corrupted or unavailable.
        In general, this means that the *tpr_file* argument should include
        the starting configuration you intend for the phase that you are about to run().
        If you are providing the *tpr_file* because you are changing parameters that
        render existing checkpoints incompatible, you need to either generate the file
        with the checkpoint from which you want to continue, or you may remove the
        checkpoint file from the phase directory and restart that phase.

        Example
        -------
        >>> config_params = {
        ...     "tpr": "{}/topol.tpr".format(data_dir),
        ...     "ensemble_num": 1,
        ...     "ensemble_dir": tmpdir,
        ...     "pairs_json": "{}/pair_data.json".format(data_dir)
        ... }
        >>> rc = RunConfig(**config_params)
        >>> assert rc.run_data.get('phase') == 'training'
        >>> rc.run(threads=2)
        >>> assert rc.run_data.get('phase') == 'convergence'
        >>> rc.run()
        >>> assert rc.run_data.get('phase') == 'production'
        >>> rc.run(tpr_file=new_tpr, max_hours=23.9)

        """
        phase = self.run_data.get('phase')

        self.__set_workdir()

        if phase == 'training':
            context = self.__train(tpr_file=tpr_file, **kwargs)
            if all(getattr(potential, 'converged', True) for potential in context.potentials):
                self.run_data.set(phase='convergence')
            else:
                raise RuntimeError(
                    'Training alpha value has not converged.')
        elif phase == 'convergence':
            context = self.__converge(tpr_file=tpr_file, **kwargs)
            # TODO(https://github.com/kassonlab/run_brer/issues/18): Investigate for robustness in the case of
            #  batch workflows and MPI-enabled GROMACS.
            if all(getattr(potential, 'stop_called', True) for potential in context.potentials):
                self.run_data.set(phase='production')
        else:
            context = self.__production(tpr_file=tpr_file, **kwargs)
            requested_production_time = self.run_data.get('production_time')
            start_time = self.run_data.get('start_time')
            end_time = self.run_data.get('end_time')
            if end_time == 0.0:
                self._logger.warning(
                    'Upgrade `brer` plugin module to avoid a bug in which BRER phase may advance '
                    'prematurely. See https://github.com/kassonlab/run_brer/issues/19')
            else:
                assert end_time > start_time
                if end_time - start_time >= requested_production_time:
                    state_dir = os.path.dirname(self.state_json)
                    # The state.json file is overwritten below.
                    # We retain previous state.json files from
                    # completed BRER iteration "i" as state_i.json.
                    # See https://github.com/kassonlab/run_brer/issues/24.
                    archive_name = 'state_' + str(self.run_data.get('iteration')) + '.json'
                    prev_iter_state_json = os.path.join(state_dir, archive_name)
                    if os.path.exists(prev_iter_state_json):
                        message = 'If you intended to re-run the phase, please remove '
                        message += str(prev_iter_state_json)
                        message += '. See https://github.com/kassonlab/run_brer/issues/24'
                        raise RuntimeError(message)
                    else:
                        shutil.copy2(self.state_json, prev_iter_state_json)
                    self.run_data.set(phase='training',
                                      start_time=0,
                                      iteration=(self.run_data.get('iteration') + 1))
        self.run_data.save_config(self.state_json)
        return context



"""Organizes the directory structure for BRER runs. Creates directories on the
fly.

How the directory structure is organized:
    * This script should be run from your "top" directory (where
      you are planning to run all your ensemble members)
    * The top directory contains ensemble member subdirectories
      This script is intended to handle just ONE ensemble member,
      so we'll only be concerned with a single member subdirectory.
    * The example below is for the first iteration (iter 0) of one
      of the members. Future iterations would go in directories
      1,2,...y

"""

import os
import pathlib


class DirectoryHelper:
    def __init__(self, top_dir, param_dict):
        """Small class for manipulating a standard directory structure for BRER
        runs.

        Parameters
        ----------
        top_dir :
            the path to the directory containing all the ensemble members.
        param_dict :
            a dictionary specifying the ensemble number, the iteration,
            and the phase of the simulation.

        Examples
        --------
        >>> .
        >>> ├── 0
        >>> │   ├── converge_dist
        >>> │   │   ├── state.cpt
        >>> │   │   ├── state_prev.cpt
        >>> │   │   └── traj_comp.part0001.xtc
        >>> │   ├── production
        >>> │   │   ├── confout.part0005.gro
        >>> │   │   ├── state.cpt
        >>> │   │   ├── state_prev.cpt
        >>> │   │   ├── state_step4622560.cpt
        >>> │   │   ├── traj_comp.part0002.xtc
        >>> │   └── training
        >>> │       ├── state.cpt
        >>> │       ├── state_prev.cpt
        >>> │       └── traj_comp.part0001.xtc
        >>> ├── state.json
        >>> ├── state_{iteration}.json
        >>> ├── submit.slurm
        >>> └── syx.tpr

        """

        self._top_dir = str(top_dir)
        self._required_parameters = ['ensemble_num', 'iteration', 'phase']
        for required in self._required_parameters:
            if required not in param_dict:
                raise ValueError('Must define {}'.format(required))
        self._param_dict = param_dict

    def get_dir(self, level: str) -> str:
        """Get the directory for however far you want to go down the directory
        tree.

        Parameters
        ----------
        level :
            one of 'top', 'ensemble_num', 'iteration', or 'phase'.
            See the directory structure example provided at the beginning of this class.

        Returns
        -------
        str
            the path to the specified directory *level*.

        """
        pdict = self._param_dict
        if level == 'top':
            return_dir = self._top_dir
        elif level == 'ensemble_num':
            return_dir = '{}/mem_{}'.format(self._top_dir, pdict['ensemble_num'])
        elif level == 'iteration':
            return_dir = '{}/mem_{}/{}'.format(self._top_dir,
                                               pdict['ensemble_num'],
                                               pdict['iteration'])
        elif level == 'phase':
            return_dir = '{}/mem_{}/{}/{}'.format(self._top_dir,
                                                  pdict['ensemble_num'],
                                                  pdict['iteration'],
                                                  pdict['phase'])
        else:
            raise ValueError('{} is not a valid directory type for BRER '
                             'simulations'.format('type'))
        return return_dir

    def build_working_dir(self):
        """Checks to see if the working directory for current state of BRER
        simulation exists. If it does not, creates the directory.
        """
        phase_dir = pathlib.Path(self.get_dir('phase')).absolute()
        os.makedirs(phase_dir, mode=0o755, exist_ok=True)

    def change_dir(self, level):
        """Change to directory specified by level.

        Parameters
        ----------
        level : str
            How far to go down the directory tree.
            Can be one of 'top', 'ensemble_num', 'iteration', or 'phase'.
        """
        next_dir = self.get_dir(level)
        os.chdir(next_dir)
        return next_dir



"""Unit tests and regression for RunData classes."""
import json
import tempfile

import pytest

from brer.pair_data import PairData
from brer.pair_data import PairDataCollection
from brer.run_data import PairParams
from brer.run_data import RunData


def test_pair_parameters(raw_pair_data):
    """Set PairParams objects from PairData objects. Check that there are no
    missing keys.

    Parameters
    ----------
    raw_pair_data : dict
        raw pair data from conftest.py


    """
    for key, data in raw_pair_data.items():
        # We removed the *name* field from input pair_data.json in 2.0.
        assert 'name' not in data
        pd = PairData(name=key, **data)
        pp = PairParams(key, sites=pd.sites)
        for default_param in ('alpha', 'target'):
            assert hasattr(pp, default_param)


def test_run_data(tmpdir, raw_pair_data):
    """Test creation of RunData object from PairData objects and check no
    missing keys.

    Parameters
    ----------
    tmpdir : str
        pytest temporary directory
    raw_pair_data : dict
        raw pair data from conftest.py


    """
    pairs = PairDataCollection(*[PairData(name=name, **pair) for name, pair in raw_pair_data.items()])
    rd = RunData.create_from(pairs)

    # Get an arbitrary but valid name.
    name = tuple(raw_pair_data.keys())[-1]
    # Test the setting function
    with pytest.raises(ValueError):
        rd.set(tau=0.1, name=name)
    with pytest.raises(ValueError):
        rd.set(alpha=1.)
    rd.set(alpha=1., name=name)

    # Test getting
    rd.get("alpha", name=name)
    with pytest.raises(ValueError):
        rd.get("alpha")

    # Test read/write of the state
    rd.save_config("{}/state.json".format(tmpdir))
    old_rd = rd
    rd = RunData.create_from("{}/state.json".format(tmpdir))

    assert old_rd.as_dictionary() != rd.as_dictionary()
    rd.set(alpha=1., name=name)
    # Confirm that the restored data is the same as the original.
    assert old_rd.as_dictionary() == rd.as_dictionary()

    with tempfile.NamedTemporaryFile(suffix='.json', mode='w') as tmp:
        test_data = raw_pair_data.copy()
        for name in test_data:
            test_data[name]['name'] = name
        json.dump(obj=test_data, fp=tmp)
        tmp.flush()
        # Tolerates redundant *name* field in the pair_data.json file,
        # but issues deprecation warning.
        with pytest.deprecated_call():
            PairDataCollection.create_from(tmp.name)

    with tempfile.NamedTemporaryFile(suffix='.json', mode='w') as tmp:
        test_data = raw_pair_data.copy()
        for name in test_data:
            test_data[name]['name'] = f'{name}_different'
        json.dump(test_data, tmp)
        tmp.flush()
        # Tolerates redundant *name* field. Warns if inconsistent.
        with pytest.warns(UserWarning, match='instead of'):
            PairDataCollection.create_from(tmp.name)



"""Compatibility support for older Python versions.

Provide some imports and alternative definitions depending on Python version
so that we don't have to have `sys.version_info` conditionals scattered around
the package.

* Built-in and :py:mod:`collections.abc` containers are not Generic before 3.9.
* Dataclasses allow tighter control in Python 3.10

"""
__all__ = ['Iterable', 'List', 'Mapping', 'MutableMapping', 'dataclass_kw_only', 'dataclass_slots']

import collections.abc
import sys
import typing

assert sys.version_info.major >= 3
_T = typing.TypeVar('_T')

if sys.version_info.major > 3 or sys.version_info.minor >= 9:
    List: typing.Generic[_T] = list
    """Generic list type."""
else:
    from typing import List

if sys.version_info.major > 3 or sys.version_info.minor >= 9:
    Iterable = collections.abc.Iterable
    Mapping = collections.abc.Mapping
    MutableMapping = collections.abc.MutableMapping
else:
    Iterable = typing.Iterable
    Mapping = typing.Mapping
    MutableMapping = typing.MutableMapping

dataclass_kw_only: collections.abc.Mapping
"""Require keyword arguments for dataclass initialization.

We are trying to reduce ambiguity by confirming that the ``name`` field is often
redundant with a key used to store such an object. To that end, we are trying
to be more mindful of how these objects are constructed. We will try to prevent
some previous usage patterns in which data structures were initialized with a
positional *name* argument, which was then overwritten by a subsequent *set()*.
Dataclass init arguments can be provided positionally in the order in which
fields are defined, unless *kw_only=True* (after Py 3.10).
"""

dataclass_slots: collections.abc.Mapping
"""Disallow adding attributes to instances.

In Python >= 3.10, the *slots* option to dataclasses will help keep fields
from accidentally getting added to instances.
"""

if sys.version_info.major > 3 or sys.version_info.minor >= 10:
    dataclass_kw_only = {'kw_only': True}
    dataclass_slots = {'slots': True}
else:
    dataclass_kw_only = {}
    dataclass_slots = {}



"""Custom PEP517 build backend.

Note: critical details of this file borrowed from
https://github.com/pybind/cmake_example/commit/31bc276d91985c9bb94e2b4ec12f3fd528971f2c

See Also:
* https://peps.python.org/pep-0517/#in-tree-build-backends
* https://setuptools.pypa.io/en/latest/build_meta.html#dynamic-build-dependencies-and-other-build-meta-tweaks

The installer should be able to find its GROMACS dependency through the installed
gmxapi Python package. If not, the installer will look for GMXTOOLCHAINDIR,
gmxapi_ROOT, and GROMACS_DIR, in that order. (These are normally defined by the
GMXRC file.) Alternatively, you can provide additional command line input to
CMake through the CMAKE_ARGS environment variable.

Help CMake find the GROMACS compiler toolchain with
``-C  /path/to/gromacs/share/cmake/gromacs{suffix}/gromacs-hints{suffix}.cmake``.
For older GROMACS installations, use
``-DCMAKE_TOOLCHAIN_FILE=/path/to/gromacs/share/cmake/gromacs{suffix}/gromacs-toolchain{suffix}.cmake``

Example:
    export GROMACS_DIR=/path/to/gromacs
    export CMAKE_ARGS="-Dgmxapi_ROOT=$GROMACS_DIR -C $GROMACS_DIR/share/cmake/gromacs/gromacs-hints.cmake"
    pip install brer-md

"""
import logging
import os
import pathlib
import re
import subprocess
import sys
import typing
import warnings

try:
    from functools import cache
except ImportError:
    # functools.cache is not available until Python 3.9
    import functools
    cache = functools.lru_cache(maxsize=None)

from setuptools import Extension
from setuptools.command.build_ext import build_ext
from setuptools import build_meta as _orig

if hasattr(str, 'removeprefix'):
    removeprefix = str.removeprefix
else:
    # str.removeprefix was added in 3.9

    def removeprefix(s: str, __prefix: str):
        if s.startswith(__prefix):
            return s[len(__prefix):]
        else:
            return s


class BrerBuildSystemError(Exception):
    """Error processing setup.py for gmxapi Python package."""


# Note that the setuptools.build_meta module objects `build_wheel`, etc, are references
# to bound methods of a _BuildMetaBackend instance that is instantiated at import
# (as _BACKEND).
#
# As of version 65, the setuptools hooks still call a `setup.py` script under the hood
# with the traditional distutils commands. `pip` and `build` will collect command line
# arguments to be provided through the `config_settings` dict (which we can also
# process or extend as we like). Only `--global-option` and `--build-option` are
# officially allowed as root level keys, and allowed global options are not officially
# extensible. See setuptools.build_meta._ConfigSettings.

# Mandatory hooks
build_wheel = _orig.build_wheel
build_sdist = _orig.build_sdist
# If we can't get _BuildMetaBackend.build_wheel() to call run_setup() with our
# custom arguments, we will need to replace part of that call stack, such as by
# subclassing _BuildMetaBackend and overriding.
# def build_wheel():
#     # Need to override _BuildMetaBackend.run_setup()
#     return _orig.build_wheel()


# Optional hooks

prepare_metadata_for_build_wheel = _orig.prepare_metadata_for_build_wheel


def get_requires_for_build_wheel(config_settings=None):
    return _orig.get_requires_for_build_wheel(config_settings)  # + [...]


def get_requires_for_build_sdist(self, config_settings=None):
    return _orig.get_requires_for_build_sdist(config_settings)  # + [...]


# PEP 660: editable installs
# Ref: https://peps.python.org/pep-0660/
if hasattr(_orig, 'LEGACY_EDITABLE') and not _orig.LEGACY_EDITABLE:
    build_editable = _orig.build_editable
    get_requires_for_build_editable = _orig.get_requires_for_build_editable
    prepare_metadata_for_build_editable = _orig.prepare_metadata_for_build_editable

# gmxapi does not officially support Windows environments.
# However, we can try to be friendly or prepare for a possible future
# in which we can support more platforms.
# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(typing.cast('setuptools.Command', build_ext)):
    """Derived distutils Command for build_extension.

    See https://github.com/pybind/cmake_example for the current version
    of the sample project from which this is borrowed.
    """

    def build_extension(self, ext):
        # The pybind11 package is only needed for `build_ext`, and may not be installed
        # in the caller's environment. By the time this function is called, though,
        # build dependencies will have been checked.
        import pybind11

        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # required for auto-detection & inclusion of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        cfg = "Debug" if debug else "Release"

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={}".format(extdir),
            "-DCMAKE_BUILD_TYPE={}".format(cfg),  # not used on MSVC, but no harm
        ]
        build_args = []
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSx on conda-forge)
        if "CMAKE_ARGS" in os.environ:
            cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]

        if self.compiler.compiler_type != "msvc":
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator:
                try:
                    import ninja  # noqa: F401

                    cmake_args += ["-GNinja"]
                except ImportError:
                    pass

        else:

            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
                ]
                build_args += ["--config", cfg]

        if sys.platform.startswith("darwin"):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += ["-j{}".format(self.parallel)]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        cmake_args = update_gromacs_client_cmake_args(cmake_args)

        has_pybind = False
        for arg in cmake_args:
            if arg.upper().startswith('-DPYBIND11_ROOT'):
                has_pybind = True
        if not has_pybind:
            pybind_root = pybind11.get_cmake_dir()
            if pybind_root:
                cmake_args.append(f'-Dpybind11_ROOT={pybind_root}')

        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


def get_gmxapi_config():
    """Get the GROMACS configuration dictionary from the gmxapi package, if available.

    Returns
    -------
    dict
        Contents vary by gmxapi version. (May be empty.)

    """
    try:
        from gmxapi.utility import config
    except ImportError:
        try:
            from gmxapi.commandline import _config as config
        except ImportError:
            config = None

    if config:
        return config()
    else:
        return {}


def _parse_cmake_defines(args: typing.Sequence[str]):
    """Handle various possible syntax for CMake ``-D`` options.

    Ref: https://cmake.org/cmake/help/latest/manual/cmake.1.html#options
    """
    for i, arg in enumerate(args):
        if arg.startswith('-D'):
            definition = removeprefix(arg, '-D')
            if not definition:
                definition = args[i+1]
            key, value = definition.split('=', maxsplit=1)
            key = key.split(':')[0]
            yield key, value


# @cache
def get_cmake_defines(args: typing.Sequence[str]):
    """Get key-value tuples for CMake variables defined as arguments.

    Ref: https://cmake.org/cmake/help/latest/manual/cmake.1.html#options
    """
    return dict(_parse_cmake_defines(args))


# @cache
def cmake_defined(key: str, args: typing.Sequence[str]):
    """Return the value of the key, if defined as a CMake argument, else None."""
    return get_cmake_defines(args).get(key, None)


def guess_prefixes(
        *,
        config: dict,
        keys: typing.Iterable[str]):
    """Generate search roots from the configuration dictionary and environment.

    Parameters
    ----------
    config : dict[str, str]
        dictionary of CMake definitions or other config values.
    keys : Iterable[str]
        environment variables or *config* entries to check.

    """
    sources = [config[key] if key in config else os.getenv(key) for key in keys]
    for source in sources:
        if source:
            # PATH-like environment variables or config values may be lists delimited by : or ;
            for _first_split in source.split(':'):
                for _second_split in _first_split.split(';'):
                    yield pathlib.Path(_second_split).resolve()


def get_search_paths(prefix: pathlib.Path, package: str):
    """Generator to produce the minimum number of search paths."""
    # Follow the search logic described at
    # https://cmake.org/cmake/help/latest/command/find_package.html#config-mode-search-procedure
    # This generator function tries to minimize the amount of filesystem interaction,
    # but is structured procedurally so that its logic can be most easily compared with
    # the heuristics documented for CMake.
    yield prefix.resolve()
    package_upper = package.upper()
    for first in [_ for _ in prefix.resolve().iterdir() if _.is_dir()]:
        # < prefix > / (cmake | CMake) /
        if first.name.upper() == 'CMAKE':
            yield first
        elif first.name.upper().startswith(package_upper):
            # < prefix > / < name > * /
            yield first
            for second in [_ for _ in first.iterdir() if _.is_dir()]:
                # < prefix > / < name > * / (cmake | CMake) /
                if second.name.upper() == 'CMAKE':
                    yield second
                    # < prefix > / < name > * / (cmake | CMake) / < name > * /
                    for third in [_ for _ in second.iterdir() if _.is_dir()]:
                        if third.name.upper().startswith(package_upper):
                            yield third
                elif second.name == 'lib':
                    logging.debug(
                        f'Skipping search for {str(second)}/<arch>. '
                        'This backend does not interpret CMAKE_LIBRARY_ARCHITECTURE. '
                        'Please provide an exact directory for the gmxapi CMake config file.')
                    continue
                elif second.name == 'share' or second.name.startswith('lib'):
                    for third in [_ for _ in second.iterdir() if _.is_dir()]:
                        if third.name == 'cmake':
                            for fourth in [_ for _ in third.iterdir() if _.is_dir()]:
                                # < prefix > / < name > * / (lib / < arch >| lib * | share) / cmake / < name > * /
                                if fourth.name.upper().startswith(package_upper):
                                    yield fourth
                        elif third.name.upper().startswith(package_upper):
                            # < prefix > / < name > * / (lib / < arch >| lib * | share) / < name > * /
                            yield third
                            # < prefix > / < name > * / (lib / < arch >| lib * | share) / < name > * / (cmake | CMake) /
                            fourth = third / 'cmake'
                            if fourth.exists() and fourth.is_dir():
                                yield fourth
                            else:
                                fourth = third / 'CMake'
                                if fourth.exists() and fourth.is_dir():
                                    yield fourth
        else:
            if first.name == 'lib':
                logging.debug(
                    f'Skipping search for {str(first)}/<arch>. '
                    'This backend does not interpret CMAKE_LIBRARY_ARCHITECTURE. '
                    'Please provide an exact directory for the gmxapi CMake config file.')
                continue
            if first.name.startswith('lib') or first.name == 'share':
                for second in [_ for _ in first.iterdir() if _.is_dir()]:
                    if second.name == 'cmake':
                        # < prefix > / (lib / < arch >| lib * | share) / cmake / < name > * /
                        for third in [_ for _ in second.iterdir() if _.is_dir()]:
                            if third.name.upper().startswith(package_upper):
                                yield third
                    elif second.name.upper().startswith(package_upper):
                        # < prefix > / (lib / < arch >| lib * | share) / < name > * /
                        yield second
                        # < prefix > / (lib / < arch >| lib * | share) / < name > * / (cmake | CMake) /
                        third = second / 'cmake'
                        if third.exists() and third.is_dir():
                            yield third
                        else:
                            third = second / 'CMake'
                            if third.exists() and third.is_dir():
                                yield third


def is_valid_package_root(*, path: typing.Optional[pathlib.Path], package: str, targets: tuple):
    if not path or not path.exists():
        return False
    for subdirectory in get_search_paths(prefix=path, package=package):
        for target in targets:
            if os.path.exists(subdirectory / target):
                return True
    return False


def package_root_from_expanded_config(
        *,
        package: str,
        targets: typing.Iterable[str],
        config: dict,
        keys: typing.Iterable[str]) -> typing.Optional[pathlib.Path]:
    """Scan the configuration dictionary and environment for a marked path.

    Parameters
    ----------
    package : str
        CMake package name.
    targets : Iterable[str]
        config file names that would indicate a valid directory.
    config : dict[str, str]
        dictionary of CMake definitions or other config values.
    keys : Iterable[str]
        environment variables or *config* entries to check.

    """
    if not isinstance(targets, tuple):
        raise ValueError('targets must be a tuple')
    for prefix in guess_prefixes(config=config, keys=keys):
        if is_valid_package_root(path=prefix, package=package, targets=targets):
            return prefix


def get_gmxapi_root(*, config: dict, args: typing.Sequence[str]):
    """Get the value for the gmxapi_ROOT CMake variable.

    First examine the *config* dictionary for hints. Then check the environment,
    """
    # directory contents that might indicate a viable gmxapi_ROOT or CMake package root.
    sentries = ('gmxapi', 'cmake', 'share', 'gmxapi-config.cmake', 'gmxapiConfig.cmake')
    path = config.get('gmxapi_root', None)
    if path:
        path = pathlib.Path(path).resolve()
    if path and not path.exists():
        warnings.warn(f'Installed gmxapi reports non-existent root path {path}.')
        path = None
    if not path:
        # Config values that might serve as hints:
        hints = ('gmx_cmake_hints', 'gmx_cmake_toolchain', 'gmx_executable', 'gmx_bindir')
        for hint in [pathlib.Path(config[key]).resolve() for key in hints if key in config]:
            guess = hint.parent
            while guess != guess.root and not path:
                for sentry in sentries:
                    if (guess / sentry).exists():
                        path = guess
                        break
                guess = guess.parent
            if path and path.exists():
                break

    # The gmxapi CMake config file is in a path that looks something like
    # $CMAKE_INSTALL_PREFIX/share/cmake/gmxapi/gmxapiConfig.cmake
    # Various gromacs components are installed to directories that are parallel
    # to some part of this path.
    # One of the following files must be discoverable in gmxapi_ROOT.
    # Ref https://cmake.org/cmake/help/latest/command/find_package.html#config-mode-search-procedure
    # Reduce globbing: These are the only two config file names possible through gmxapi 0.4.
    targets = ('gmxapi-config.cmake', 'gmxapiConfig.cmake')
    # Check the environment. Let CMake defines override environment variables.
    cmake_vars = get_cmake_defines(args)
    if not path or not is_valid_package_root(path=path, targets=targets, package='gmxapi'):
        path = package_root_from_expanded_config(
            package='gmxapi',
            targets=targets,
            config=cmake_vars,
            keys=('gmxapi_ROOT', 'GMXAPI_ROOT', 'GMXAPI_DIR', 'GROMACS_DIR', 'CMAKE_PREFIX_PATH'))

    # One more try...
    if not path or not is_valid_package_root(path=path, targets=targets, package='gmxapi'):
        # Try to guess from args.
        hint = ''
        if '-C' in args:
            hint = args[args.index('-C') + 1]
        if not hint:
            for guess in ('CMAKE_TOOLCHAIN_FILE', 'GMXTOOLCHAINDIR'):
                if guess in cmake_vars:
                    hint = cmake_vars[guess]
                else:
                    hint = os.getenv(guess)
                if hint:
                    break
        if hint and os.path.exists(hint):
            guess = pathlib.Path(hint).resolve()
            while guess != guess.root and not path:
                for sentry in sentries:
                    if (guess / sentry).exists():
                        path = guess
                        break
                guess = guess.parent
    if path:
        return str(path)
    else:
        return None


def get_cmake_hints(*, config: dict, args: typing.Sequence[str]):
    """Get a file for the -C CMake argument, if possible.

    Check the *config* dict, first. If -C is provided in *args*, use the
    user-provided path, but warn if overriding a path from *config*.

    If a hints file is not provided explicitly, try to locate a hints file from
    other values in *config* or *args*.

    Ref: https://cmake.org/cmake/help/latest/manual/cmake.1.html#options
    """
    hints_file = config.get('gmx_cmake_hints', None)
    if '-C' in args:
        hint = args[args.index('-C') + 1]
        if os.path.exists(hint):
            hint = pathlib.Path(hint).resolve()
            if hints_file is not None and not hint.samefile(hints_file):
                warnings.warn(
                    f'Overriding gmxapi configuration (hints file {hints_file}) with "-C {hint}"'
                )
            hints_file = str(hint)
    if hints_file and not os.path.exists(hints_file):
        raise BrerBuildSystemError(
            f'"-C {hints_file}" refers to non-existent CMake initial-cache file.'
        )
    return hints_file


def get_toolchain_file(*, config: dict, args: typing.Sequence[str]):
    toolchain_file = os.getenv('CMAKE_TOOLCHAIN_FILE')
    if not toolchain_file:
        toolchain_file = config.get('gmx_cmake_toolchain', None)
    if not toolchain_file:
        suffix = os.getenv('GROMACS_SUFFIX', cmake_defined('GROMACS_SUFFIX', args))
        if suffix is None:
            suffix = config.get('gmx_suffix', None)
        if suffix is None:
            if 'gmx_mpi_type' in config and 'gmx_double' in config:
                suffix = ''
                if config['gmx_mpi_type'] == 'library':
                    suffix += '_mpi'
                if config['gmx_double']:
                    suffix += '_d'
        if suffix is None:
            # Guessing the suffix is too hard on older GROMACS/gmxapi combinations.
            # See https://gitlab.com/gromacs/gromacs/-/issues/4334
            return None
        gmx_toolchain_dir = os.getenv('GMXTOOLCHAINDIR', cmake_defined('GMXTOOLCHAINDIR', args))
        if gmx_toolchain_dir and os.path.exists(gmx_toolchain_dir):
            toolchain_file = os.path.join(
                gmx_toolchain_dir,
                f'gromacs{suffix}',
                'gromacs-toolchain' + suffix + '.cmake')
    return toolchain_file


def update_gromacs_client_cmake_args(args: typing.Sequence[str]) -> typing.List[str]:
    """Try to convert information from command line environment to usable client CMake stuff.

    Normalize user input and automated GROMACS detection to produce a list of CMake arguments
    containing a ``-Dgmxapi_ROOT`` and, if possible, hints for generating the CMake build tool
    chain.

    Args:
        args: CMake args provided by CMakeBuild instance, including user input from CMAKE_ARGS.

    First, we must determine a single value of ``gmxapi_ROOT``. If available, we
    first check `gmxapi.utils.config()` or `gmxapi.commandline._config()` for as
    much information as we can derive. We allow user-provided information to
    override the gmxapi package config with a warning.

    If ``CMAKE_ARGS`` contains a ``-C`` option, the value is checked for consistency with the
    ``gmxapi_ROOT`` path.

    If ``CMAKE_ARGS`` contains both ``-C`` and ``-DCMAKE_TOOLCHAIN_FILE`` options,
    both are passed along to CMake and we assume the user knows what they are doing.

    If ``CMAKE_ARGS`` contains neither a ``-C`` option, nor a ``-DCMAKE_TOOLCHAIN_FILE`` option,
    this script attempts to locate a ``gromacs-hints.cmake`` file in order to generate
    a ``-C`` option to add to the CMake arguments for CMakeBuild.
    If the script is unable to locate a ``gromacs-hints.cmake`` file, we fall back
    to the gmxapi 0.2 scheme for ``GMXTOOLCHAINDIR``.

    This function compartmentalizes details that are likely to evolve with issues
    https://gitlab.com/gromacs/gromacs/-/issues/3273
    and
    https://gitlab.com/gromacs/gromacs/-/issues/3279

    See linked issues for more discussion or to join in the conversation.

    This logic can be simplified as support is dropped for older GROMACS or
    gmxapi versions.
    """
    args = list(str(arg) for arg in args)

    gmxapi_config = get_gmxapi_config()
    gmxapi_root = get_gmxapi_root(config=gmxapi_config, args=args)
    if gmxapi_root and cmake_defined('gmxapi_ROOT', args) is None:
        args.append(f'-Dgmxapi_ROOT={gmxapi_root}')

    cmake_hints = get_cmake_hints(config=gmxapi_config, args=args)
    if cmake_hints and '-C' not in args:
        args.extend(('-C', cmake_hints))

    if '-C' not in args:
        # Try to derive a toolchain file argument, if not already present.
        toolchain_file = cmake_defined('CMAKE_TOOLCHAIN_FILE', args)
        if not toolchain_file:
            toolchain_file = get_toolchain_file(config=gmxapi_config, args=args)
            if toolchain_file and os.path.exists(toolchain_file):
                args.append(f'-DCMAKE_TOOLCHAIN_FILE={toolchain_file}')

    return args



import contextlib
import json
import logging
import os
import shutil
import tempfile

import pytest

from brer.run_config import RunConfig

logger = logging.getLogger()

try:
    from mpi4py import MPI
except ImportError:
    MPI = None

try:
    from gmxapi.utility import config
except ImportError:
    config = dict

# To use PyCharm debug server, you would need something like the following:
# if rank == 0:
#     import pydevd_pycharm
#
#     pydevd_pycharm.settrace('localhost',
#                             port=33333,
#                             stdoutToServer=True,
#                             stderrToServer=True)

with_mpi_only = pytest.mark.skipif(
    MPI is None or MPI.COMM_WORLD.Get_size() < 2,
    reason='This test requires mpi4py and a usable MPI environment.')

# Try to get a reasonable number of threads to use.
try:
    num_cpus: int = len(os.sched_getaffinity(0))
except Exception:
    num_cpus = 4


def test_import_utility(simulation_input):
    from brer.run_config import _context, from_tpr, WorkElement, get_api_callable
    _context()
    element = from_tpr(simulation_input)
    assert isinstance(element, WorkElement)
    func = get_api_callable("nonsense", ("missing1",))
    with pytest.raises(RuntimeError, match="Could not import missing1"):
        func()
    func = get_api_callable("nonsense", ("missing1", "missing2"))
    with pytest.raises(RuntimeError, match="Could not import missing2"):
        func()
    with pytest.raises(RuntimeError, match="Could not import missing1"):
        func()


@contextlib.contextmanager
def working_directory_fence():
    """Ensure restoration of working directory when leaving context manager."""
    wd = os.getcwd()
    try:
        yield wd
    finally:
        os.chdir(wd)


def test_run_config(tmpdir, pair_data_file, simulation_input):
    with working_directory_fence():
        config_params = {
            "tpr": simulation_input,
            "ensemble_num": 1,
            "ensemble_dir": tmpdir,
            "pairs_json": pair_data_file
        }
        os.makedirs(
            "{}/mem_{}".format(tmpdir, config_params["ensemble_num"]),
            exist_ok=True
        )
        rc = RunConfig(**config_params)
        rc.run_data.set(A=5,
                        tau=0.1,
                        tolerance=0.1,
                        num_samples=2,
                        sample_period=0.001,
                        production_time=10000.)

        # Training phase.
        assert rc.run_data.get('iteration') == 0
        assert rc.run_data.get('phase') == 'training'

        # With a tight tolerance and short run time, it can't have converged.
        # See https://github.com/kassonlab/run_brer/issues/8
        with pytest.raises(RuntimeError):
            rc.run(max_hours=0.0001)
        assert rc.run_data.get('phase') == 'training'

        # Allow the training phase to converge.
        rc.run_data.set(tolerance=10000)

        # Include a test for kwarg handling.
        kwargs = dict()
        # From gmxapi/CMakeLists.txt, this is "library", "tmpi", or None
        mpi_type = config().get("gmx_mpi_type")
        if mpi_type == "tmpi":
            kwargs["threads"] = num_cpus
        rc.run(**kwargs)

        # Convergence phase.
        assert rc.run_data.get('phase') == 'convergence'
        # Check that bad alpha is caught.
        _original_alpha = [None] * len(rc.run_data.pair_params)
        for i, name in enumerate(rc.run_data.pair_params):
            _original_alpha[i] = rc.run_data.get('alpha', name=name)
            rc.run_data.set(name=name, alpha=0.0)
        with pytest.raises(ValueError):
            rc.run()
        # Set alpha back so we can continue.
        for i, name in enumerate(rc.run_data.pair_params):
            rc.run_data.set(name=name, alpha=_original_alpha[i])
        # Check robustness to early termination.
        rc.run_data.set(tolerance=0.01)
        rc.run(max_hours=0.001)
        assert rc.run_data.get('phase') == 'convergence'

        # Finish convergence phase.
        rc.run_data.set(tolerance=10000)
        rc.run()

        # Production phase.
        assert rc.run_data.get('phase') == 'production'
        with pytest.raises(TypeError):
            # Test handling of kwarg collisions.
            rc.run(end_time=1.0)
        # Note that rc.__production failed, but rc.run() will have changed directory.
        # This is an unspecified side effect, but we can use it for some additional
        # inspection.
        assert len(os.listdir()) == 0
        # Test another kwarg.
        rc.run(max_hours=0.001)
        assert rc.run_data.get('phase') == 'production'

        # Test the production phase bootstrapping option.
        # It is a little difficult to test that the production phase actually
        # runs with a non-default TPR file.
        # Warning: This may need some extra conditional logic to support more gmxapi
        # versions.
        with tempfile.TemporaryDirectory() as directory:
            new_tpr = os.path.join(directory, 'tmp.tpr')
            shutil.copy(simulation_input, new_tpr)
            gmxapi_context = rc.run(tpr_file=new_tpr, max_hours=0.001)
        element = json.loads(gmxapi_context.work.elements['tpr_input'])
        assert str(element['params']['input'][0]) == str(new_tpr)
        assert rc.run_data.get('phase') == 'production'
        assert rc.run_data.get('iteration') == 0


@with_mpi_only
def test_mpi_ensemble(tmpdir, pair_data_file, simulation_input):
    """Test a batch of multiple ensemble members in a single MPI context."""
    with working_directory_fence():
        comm: MPI.Comm = MPI.COMM_WORLD
        rank: int = comm.Get_rank()
        logger.info(f'rank is {rank}')
        assert rank < comm.Get_size()

        tpr_list = [simulation_input] * comm.Get_size()
        config_params = {
            "tpr": tpr_list,
            "ensemble_num": None,
            "ensemble_dir": tmpdir,
            "pairs_json": pair_data_file
        }
        os.makedirs(
            "{}/mem_{}".format(tmpdir, config_params["ensemble_num"]),
            exist_ok=True
        )
        rc = RunConfig(**config_params)
        rc.run_data.set(A=5,
                        tau=0.1,
                        tolerance=100,
                        num_samples=2,
                        sample_period=0.1,
                        production_time=0.2)

        # Training phase.
        assert rc.run_data.get('phase') == 'training'
        # Include a test for kwarg handling.
        kwargs = dict()
        mpi_type = config().get("gmx_mpi_type")
        if mpi_type == "tmpi":
            kwargs["threads"] = 2
        rc.run(**kwargs)

        # Convergence phase.
        assert rc.run_data.get('phase') == 'convergence'
        if mpi_type == "tmpi":
            kwargs['threads'] = 4
        rc.run(**kwargs)

        # Production phase.
        assert rc.run_data.get('phase') == 'production'
        with pytest.raises(TypeError):
            # Test handling of kwarg collisions.
            rc.run(end_time=1.0)
        # Note that rc.__production failed, but rc.run() will have changed directory.
        # This is an unspecified side effect, but we can use it for some additional
        # inspection.
        assert len(os.listdir()) == 0
        # Test another kwarg.
        kwargs['max_hours'] = 0.001
        rc.run(**kwargs)

        if comm.Get_size() > 1:
            # TODO(https://github.com/kassonlab/run_brer/issues/7): Confirm that we actually ran different ensemble members.
            ...



"""Handle simulation data for BRER simulations.

`RunData` manages general parameters (`GeneralParams`) and pair-specific
parameters (`PairParams`) for a single simulator for a specific phase of the
BRER method.

Parameters are initially provided through the
:py:class:`~brer.run_config.RunConfig`, and are then stored to (and restored
from) an internally managed :file:`state.json` file.

Not all parameters are applicable to all BRER phases.

See Also
--------
:py:mod:`brer.plugin_configs`

"""
import collections.abc
import dataclasses
import json
import os
import pathlib
import typing
import warnings
from dataclasses import dataclass
from dataclasses import field

from ._compat import dataclass_slots
from ._compat import List
from ._compat import Mapping
from ._compat import MutableMapping
from .pair_data import PairDataCollection


@dataclass(**dataclass_slots)
class GeneralParams:
    """Store the parameters shared by all restraints in a single simulation.

    These include some of the "Voth" parameters: *tau*, *A*, *tolerance*

    .. versionadded:: 2.0
        The *end_time* parameter.

    .. versionchanged:: 2.0
        *ensemble_num* now defaults to 0 for consistency with :py:class:`~brer.run_config.RunConfig`

    Update general parameters before a call to
    :py:meth:`brer.run_config.RunConfig.run()` by calling
    :py:meth:`brer.run_data.RunData.set()` without a *name* argument.
    """
    name: str = field(init=False, default='general')
    A: float = 50.
    end_time: float = 0.
    ensemble_num: int = 0
    iteration: int = 0
    num_samples: int = 50
    phase: str = 'training'
    production_time: float = 10000.
    sample_period: float = 100.
    start_time: float = 0.
    tau: float = 50
    tolerance: float = 0.25


@dataclass(**dataclass_slots)
class PairParams:
    """Stores the parameters that are unique to a specific restraint.

    *PairParams* is a mutable structure for run time data. Fields such as
    *alpha* and *target* may be updated automatically while running *brer*.

    *PairParams* should not be confused with
    :py:class:`~brer.pair_data.PairData` (an input data structure).

    Update pair-specific parameters before a call to
    :py:meth:`brer.run_config.RunConfig.run()` by calling
    :py:meth:`brer.run_data.RunData.set()`, providing the pair name with the
    *name* argument.

    *logging_filename* is derived from the pair *name* (user-provided;
    usually derived from the residue IDs defining the pair).
    Overriding the default produces a warning.

    .. versionchanged:: 2.0

        *sites* is required to initialize the object.


    """
    name: str
    sites: List[int]
    logging_filename: str = None
    alpha: float = 0.
    # TODO: Are we sure we should have defaults, especially for `target`?
    target: float = 3.

    def __post_init__(self):
        logging_filename = f'{self.name}.log'
        if self.logging_filename and self.logging_filename != logging_filename:
            warnings.warn(
                f'Specified logging filename {self.logging_filename} overrides default ("'
                f'{logging_filename}")')
        else:
            self.logging_filename = logging_filename


class RunData:
    """Store (and manipulate, to a lesser extent) all the metadata for a BRER run.

    The full set of metadata for a single BRER run includes both the general
    parameters and the pair-specific parameters.

    Key-value pairs provided to *general_params* will be used to update the
    default values of a new :py:class:`GeneralParams` instance.

    *pair_params* is a mapping of named `PairParams` instances.

    Both general and pair-specific parameters may be updated with *set()*.

    This is the BRER program state data structure. We avoid the name "state"
    because of potential confusion with concepts like energetic, conformational,
    or thermodynamic state, but we use the filename :file:`state.json` for the
    serialized object. RunData instances can be serialized to a file with
    :py:meth:`~brer.run_data.RunData.save_config()` or deserialized (restored from a file)
    with :py:meth:`~brer.run_data.RunData.create_from()`.


    Examples
    --------
    ::

        ├── pair parameters
        │   ├── name of pair 1
        │   │   ├── alpha
        │   │   ├── target
        │   │   └── ...
        │   ├── name of pair 2
        |
        ├── general parameters
            ├── A
            ├── tau
            ├── ...


    """
    general_params: GeneralParams
    pair_params: MutableMapping[str, PairParams]

    def __init__(self, *,
                 general_params: GeneralParams,
                 pair_params: Mapping[str, PairParams]):
        if not isinstance(general_params, GeneralParams):
            raise ValueError('*general_params* must be a GeneralParams instance.')
        if not all(isinstance(obj, PairParams) for obj in pair_params.values()):
            raise ValueError('*pair_params* must be a collection of PairParams instances.')

        self.pair_params = dict()
        for name, obj in pair_params.items():
            replacements = dict()
            if not obj.name:
                replacements[name] = name
            elif obj.name != name:
                raise ValueError(f'*pair_params* key {name} does not match *name* field in {obj}.')
            self.pair_params[name] = dataclasses.replace(obj, **replacements)

        self.general_params = general_params

    def set(self, name=None, **kwargs):
        """Set either general or pair-specific parameters.

        When a *name* argument is present, sets pair-specific parameters for
        the named restraint.

        When *name* is not provided, sets general parameters.

        Parameters
        ----------
        name : str, default=None
            Restraint name, as used in the `brer.run_config.RunConfig`.

        Raises
        ------
        ValueError
            if you provide a name and try to set a general parameter or
            don't provide a name and try to set a pair-specific parameter.
        """
        if len(kwargs) == 0:
            raise TypeError(
                f'Invalid signature: {self.__class__.__qualname__}.set() called without naming any '
                f'parameters.')

        for key, value in kwargs.items():
            # If a restraint name is not specified, it is assumed that the parameter is
            # a "general" parameter.
            if not name:
                if any(field.name == key for field in dataclasses.fields(self.general_params)):
                    setattr(self.general_params, key, value)
                    continue
                else:
                    if any(field.name == key for field in dataclasses.fields(PairParams)):
                        raise ValueError(
                            f'You must provide pair *name* for which to set pair parameter {key}.')
            else:
                if any(field.name == key for field in dataclasses.fields(self.pair_params[name])):
                    setattr(self.pair_params[name], key, value)
                    continue
                else:
                    if any(field.name == key for field in dataclasses.fields(GeneralParams)):
                        raise ValueError(
                            f'{key} is a general parameter but you have provided a pair name.')
            raise ValueError(f'{key} is not a valid parameter.')

    def get(self, key, *, name=None):
        """Get either a general or a pair-specific parameter.

        Parameters
        ----------
        key : str
            The parameter to get.
        name : str, default=None
            If getting a pair-specific parameter, specify the restraint name.

        Returns
        -------
        typing.Any
            the parameter value.


        """
        if name:
            return getattr(self.pair_params[name], key)
        else:
            try:
                return getattr(self.general_params, key)
            except AttributeError as e:
                if hasattr(PairParams, key):
                    raise ValueError(
                        f'Must specify pair *name* for parameter {key}.')
                else:
                    raise e

    def as_dictionary(self):
        """Get the run metadata as a hierarchical dictionary.

        Returns
        -------
        dict
            hierarchical dictionary of metadata


        For historical reasons, the top level dictionary keys are not exact string
        matches for the object attributes.

        """
        pair_param_dict = dict(
            (name, dataclasses.asdict(pair_params)) for name, pair_params in self.pair_params.items())

        return {
            'general parameters': dataclasses.asdict(self.general_params),
            'pair parameters': dict((key, value) for key, value in pair_param_dict.items() if key != 'name')
        }

    def save_config(self, fnm='state.json'):
        """Saves the run parameters to a log file.

        Parameters
        ----------
        fnm : str, default='state.json'
            Log file for state parameters.
        """
        with open(fnm, 'w') as fh:
            json.dump(self.as_dictionary(), fh, indent=4)

    @typing.overload
    @classmethod
    def create_from(cls,
                    source: typing.Union[str, os.PathLike, pathlib.Path],
                    ensemble_num: int = None) -> 'RunData':
        ...

    @typing.overload
    @classmethod
    def create_from(cls, source: Mapping[str, dict], ensemble_num: int = None) -> 'RunData':
        ...

    @typing.overload
    @classmethod
    def create_from(cls, source: PairDataCollection, ensemble_num: int = None) -> 'RunData':
        ...

    @classmethod
    def create_from(cls, source, *, ensemble_num: int = None):
        """Create a new instance from provided data.

        Warns if *ensemble_num* is specified but contradicts *source*.
        If *ensemble_num* is not specified and is not found in *source*, the
        default value is determined by :py:class:`GeneralParams`.

        *source* is usually either a :file:`state.json` file or a

        Parameters
        ----------
        source :
            File or Python objects from which to initialize RunData.
        ensemble_num :
            Member index in the ensemble (if any).

        """
        if isinstance(source, (str, os.PathLike, pathlib.Path)):
            if not os.path.exists(source):
                raise ValueError(f'Source file not found: {source}')
            else:
                with open(source, 'r') as fh:
                    data = json.load(fh)
                return cls.create_from(source=data, ensemble_num=ensemble_num)
        elif isinstance(source, PairDataCollection):
            # PairParams comes from PairData names and sites and default values.
            pair_params = {name: PairParams(name=name, sites=pair.sites) for name, pair in source.items()}
            # GeneralParams comes from defaults
            run_data = cls(general_params=GeneralParams(), pair_params=pair_params)
            if ensemble_num is not None:
                run_data.set(ensemble_num=ensemble_num)
            return run_data
        elif isinstance(source, collections.abc.Mapping):
            # Check for *ensemble_num* agreement.
            general_params = source['general parameters']
            if 'name' in general_params:
                del general_params['name']
            general_params = GeneralParams(**general_params)
            if ensemble_num is not None:
                _source_id = general_params.ensemble_num
                if ensemble_num != _source_id:
                    warnings.warn(f'Caller provided ensemble_num={ensemble_num} overrides {_source_id} '
                                  f'from {source}.')
                    general_params.ensemble_num = ensemble_num
            pair_params = {name: PairParams(name=name, sites=fields['sites']) for name, fields in
                           source['pair parameters'].items()}
            return RunData(general_params=general_params, pair_params=pair_params)
        else:
            raise ValueError(f'{source} is not a valid source of RunData.')



"""A set of scripts for running BRER simulations using gmxapi.

Online documentation is hosted at https://kassonlab.github.io/brer_md/.

See Also
--------
Hays, J. M., Cafiso, D. S., & Kasson, P. M.
Hybrid Refinement of Heterogeneous Conformational Ensembles using Spectroscopic Data.
*The Journal of Physical Chemistry Letters.*
DOI: `10.1021/acs.jpclett.9b01407 <https://pubs.acs.org/doi/10.1021/acs.jpclett.9b01407>`__
"""
__all__ = ['__version__']

from ._version import __version__



"""Classes used to build gmxapi plugins for all phases of a BRER iteration.

Each class corresponds to one restraint potential provided by :py:mod`brer.md`.
Each instance corresponds to one restrained pair in a single simulation phase.
"""
import dataclasses
import typing
from abc import ABC
from abc import abstractmethod
from functools import singledispatchmethod

from ._compat import dataclass_kw_only
from ._compat import List


def _get_workelement() -> typing.Type:
    try:
        # noinspection PyUnresolvedReferences
        from gmxapi.simulation.workflow import WorkElement
    except (ImportError, ModuleNotFoundError):
        # Fall-back for gmxapi < 0.1.0
        try:
            # noinspection PyUnresolvedReferences
            from gmx.workflow import WorkElement
        except (ImportError, ModuleNotFoundError):
            WorkElement = None
    if WorkElement is None:
        raise RuntimeError('brer requires gmxapi. See https://github.com/kassonlab/brer_md#requirements')
    return WorkElement


@dataclasses.dataclass
class PluginConfig(ABC):
    """Abstract class for the BRER potential configurations.

    Provide base class functionality and required interface to build ``training``,
    ``convergence``, and ``production`` phase pluggable MD potentials.
    """
    name: str = dataclasses.field(init=False, **dataclass_kw_only)
    sites: List[int] = dataclasses.field(**dataclass_kw_only)
    logging_filename: str = dataclasses.field(**dataclass_kw_only)

    @abstractmethod
    def build_plugin(self):
        """Abstract method for building a plugin.

        To be determined by the phase of the simulation (training,
        convergence, production)
        """
        pass

    @singledispatchmethod
    @classmethod
    def create_from(cls, obj):
        """Get an instance of the appropriate type from the provided data structure.

        Source *obj* is either a :py:class:`~collections.abc.Mapping` or an object
        with the fields necessary to initialize a *cls* instance.

        Extra fields in *obj* are ignored.
        """
        if cls is PluginConfig:
            raise NotImplementedError
        return cls(**{key: getattr(obj, key) for key in [field.name for field in dataclasses.fields(cls) if
                                                         field.init] if hasattr(obj, key)})

    @create_from.register(dict)
    @classmethod
    def _(cls, obj: dict):
        return cls(**{key: obj[key] for key in [field.name for field in dataclasses.fields(cls) if
                                                field.init] if key in obj})


@dataclasses.dataclass
class TrainingPluginConfig(PluginConfig):
    """Configure BRER potential for the training phase.

    The BRER training phase uses the MD potential provided by :py:func:`brer.brer_restraint`.

    See https://pubs.acs.org/doi/10.1021/acs.jpclett.9b01407 for details.
    """
    name: str = dataclasses.field(init=False, default='training', **dataclass_kw_only)

    A: float = dataclasses.field(**dataclass_kw_only)
    num_samples: int = dataclasses.field(**dataclass_kw_only)
    target: float = dataclasses.field(**dataclass_kw_only)
    tau: float = dataclasses.field(**dataclass_kw_only)
    tolerance: float = dataclasses.field(**dataclass_kw_only)

    def build_plugin(self):
        """Builds training phase plugin for BRER simulations.

        Returns
        -------
        WorkElement
            a gmxapi WorkElement to be added to the workflow graph
        """
        WorkElement = _get_workelement()
        potential = WorkElement(
            namespace="brer.md",
            operation="brer_restraint",
            depends=[],
            params=dataclasses.asdict(self)
        )
        potential.name = str(self.sites)
        return potential


@dataclasses.dataclass
class ConvergencePluginConfig(PluginConfig):
    """Configure BRER potential for the convergence phase.

    The BRER convergence phase uses the MD potential provided by :py:func:`brer.linearstop_restraint`.

    See https://pubs.acs.org/doi/10.1021/acs.jpclett.9b01407 for details.
    """
    name: str = dataclasses.field(init=False, default='convergence', **dataclass_kw_only)

    alpha: float = dataclasses.field(**dataclass_kw_only)
    sample_period: float = dataclasses.field(**dataclass_kw_only)
    target: float = dataclasses.field(**dataclass_kw_only)
    tolerance: float = dataclasses.field(**dataclass_kw_only)

    def build_plugin(self):
        """Builds convergence phase plugin for BRER simulations.

        Returns
        -------
        WorkElement
            a gmxapi WorkElement to be added to the workflow graph

        Raises
        ------
        ValueError
            If inappropriate values are detected for any fields.
        """
        if self.alpha == 0.0:
            raise ValueError('Read a non-sensical alpha value: 0.0')
        WorkElement = _get_workelement()
        potential = WorkElement(
            namespace="brer.md",
            operation="linearstop_restraint",
            depends=[],
            params=dataclasses.asdict(self))
        potential.name = str(self.sites)
        return potential


@dataclasses.dataclass
class ProductionPluginConfig(PluginConfig):
    """Configure BRER potential for the convergence phase.

    The BRER production phase uses the MD potential provided by :py:func:`brer.linear_restraint`.

    See https://pubs.acs.org/doi/10.1021/acs.jpclett.9b01407 for details.
    """
    name: str = dataclasses.field(init=False, default='production', **dataclass_kw_only)

    alpha: float = dataclasses.field(**dataclass_kw_only)
    sample_period: float = dataclasses.field(**dataclass_kw_only)
    target: float = dataclasses.field(**dataclass_kw_only)

    def build_plugin(self):
        """Builds production phase plugin for BRER simulations.

        Returns
        -------
        WorkElement
            a gmxapi WorkElement to be added to the workflow graph

        Raises
        ------
        ValueError
            If inappropriate values are detected for any fields.
        """
        if self.alpha == 0.0:
            raise ValueError('Read a non-sensical alpha value: 0.0')
        WorkElement = _get_workelement()
        potential = WorkElement(
            namespace="brer.md",
            operation="linear_restraint",
            depends=[],
            params=dataclasses.asdict(self))
        potential.name = str(self.sites)
        return potential



"""Python setuptools script for custom build_ext.

This script is not intended as a direct entry point. Instead of `python setup.py ...`,
please use `pip install ...`

Reference:
https://setuptools.pypa.io/en/latest/userguide/ext_modules.html#building-extension-modules
"""

from setuptools import setup


try:
    # from _custom_build.backend import CMakeBuild, CMakeExtension
    from backend import CMakeBuild, CMakeExtension
except ImportError as e:
    raise RuntimeError(
        'Could not load in-tree build backend. See installation instructions.'
    ) from e


setup(
    ext_modules=[CMakeExtension(name='brer.md', sourcedir='src/plugin')],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False
)



"""Unit tests and regression for PairData classes."""
import dataclasses
import json

from brer.pair_data import PairData
from brer.pair_data import PairDataCollection
from brer.pair_data import sample_all


def test_pair_data_collection(raw_pair_data, pair_data_file):
    """Ensure correct structure and typing of PairData collections."""
    with open(pair_data_file, 'r') as fh:
        pairs = PairDataCollection(*(PairData(name=name, **obj) for name, obj in json.load(fh).items()))

    for name in pairs:
        assert isinstance(pairs[name], PairData)

    assert set(pairs.keys()) == set(raw_pair_data.keys())
    for pair_name, pair_data in pairs.items():
        for key, value in dataclasses.asdict(pair_data).items():
            if key != 'name':
                assert value == raw_pair_data[pair_name][key]
        for key, value in raw_pair_data[pair_name].items():
            assert getattr(pair_data, key) == value

    assert pairs == PairDataCollection.create_from(pair_data_file)


def test_pair_data_helpers(pair_data_file):
    """Test the module functions for accessing PairData and PairDataCollection."""
    pairs = PairDataCollection.create_from(pair_data_file)
    samples = sample_all(pairs)
    assert set(samples.keys()) == set(pairs.keys())
    for pair in pairs:
        assert samples[pair] >= min(pairs[pair].bins)
        assert samples[pair] <= max(pairs[pair].bins)



"""Unit tests and regression for PluginConfig classes."""
import dataclasses

import pytest

from brer.plugin_configs import ConvergencePluginConfig
from brer.plugin_configs import ProductionPluginConfig
from brer.plugin_configs import TrainingPluginConfig
from brer.run_data import GeneralParams
from brer.run_data import PairParams


def test_plugins(raw_pair_data):
    """Build all three test plugins and check that they have all the required
    data to run.

    Parameters
    ----------
    raw_pair_data : dict
        A set of pair-specific parameters. Provided in conftest.py
    """
    tpc: TrainingPluginConfig
    cpc: ConvergencePluginConfig
    ppc: ProductionPluginConfig

    general_parameter_defaults = dataclasses.asdict(GeneralParams())
    name = list(raw_pair_data.keys())[0]
    sites = raw_pair_data[name]["sites"]
    pp = PairParams(name=name, sites=sites)

    param_dict = general_parameter_defaults.copy()
    param_dict.update(**dataclasses.asdict(pp))

    # 'target' is a run-time parameter, and needs to be provided.
    # Here we choose an arbitrary value.
    param_dict['target'] = 1.0

    tpc = TrainingPluginConfig.create_from(param_dict)

    # 'alpha' is a required parameter for convergence and production phases.
    param_dict['alpha'] = 0.0
    cpc = ConvergencePluginConfig.create_from(param_dict)
    ppc = ProductionPluginConfig.create_from(param_dict)

    tpc.build_plugin()

    # By default, alpha is 0.0, which doesn't make sense in the convergence and production phases.
    assert cpc.alpha == 0.0
    with pytest.raises(ValueError):
        cpc.build_plugin()
    with pytest.raises(ValueError):
        ppc.build_plugin()

    cpc.alpha = 1.0
    cpc.build_plugin()

    ppc.alpha = 1.0
    ppc.build_plugin()
